@@ -27,18 +27,10 @@
 
 #include <osmocom/sigtran/sccp_helpers.h>
 
-#include <osmocom/ranap/ranap_common_ran.h>
-
-#if ENABLE_PFCP
-#include <osmocom/pfcp/pfcp_cp_peer.h>
-#endif
-
 #include <osmocom/hnbgw/hnbgw_cn.h>
 #include <osmocom/hnbgw/context_map.h>
+#include <osmocom/hnbgw/hnbgw_ranap.h>
 #include <osmocom/hnbgw/tdefs.h>
-#include <osmocom/hnbgw/mgw_fsm.h>
-#include <osmocom/hnbgw/ps_rab_ass_fsm.h>
-#include <osmocom/hnbgw/kpi.h>
 
 enum map_sccp_fsm_state {
 	MAP_SCCP_ST_INIT,
@@ -88,7 +80,7 @@ void map_sccp_fsm_alloc(struct hnbgw_context_map *map)
 
 	OSMO_ASSERT(map->sccp_fi == NULL);
 	map->sccp_fi = fi;
-
+	INIT_LLIST_HEAD(&map->sccp_fi_ctx.wait_cc_tx_msg_list);
 	/* trigger the timeout */
 	map_sccp_fsm_state_chg(MAP_SCCP_ST_INIT);
 }
@@ -187,16 +179,9 @@ static int tx_sccp_rlsd(struct osmo_fsm_inst *fi)
 	return osmo_sccp_tx_disconn(map->cnlink->hnbgw_sccp_user->sccp_user, map->scu_conn_id, NULL, 0);
 }
 
-static int destruct_ranap_ran_rx_co_ies(ranap_message *ranap_message_p)
-{
-	ranap_ran_rx_co_free(ranap_message_p);
-	return 0;
-}
-
 static int handle_rx_sccp(struct osmo_fsm_inst *fi, struct msgb *ranap_msg)
 {
 	struct hnbgw_context_map *map = fi->priv;
-	int rc;
 
 	/* If the FSM instance has already terminated, don't dispatch anything. */
 	if (fi->proc.terminating)
@@ -206,65 +191,25 @@ static int handle_rx_sccp(struct osmo_fsm_inst *fi, struct msgb *ranap_msg)
 	if (!msg_has_l2_data(ranap_msg))
 		return 0;
 
-	/* See if it is a RAB Assignment Request message from SCCP to RUA, where we need to change the user plane
-	 * information, for RTP mapping via MGW, or GTP mapping via UPF. */
-	ranap_message *message;
-	message = talloc_zero(OTC_SELECT, ranap_message);
-	rc = ranap_ran_rx_co_decode(message, message, msgb_l2(ranap_msg), msgb_l2len(ranap_msg));
-	if (rc == 0) {
-		talloc_set_destructor(message, destruct_ranap_ran_rx_co_ies);
+	return hnbgw_ranap_rx_data_dl(map, ranap_msg);
+}
 
-		LOGPFSML(fi, LOGL_DEBUG, "rx from SCCP: RANAP %s\n",
-			 get_value_string(ranap_procedure_code_vals, message->procedureCode));
+static void wait_cc_tx_msg_list_enqueue(struct hnbgw_context_map *map, struct msgb *ranap_msg)
+{
+	talloc_steal(map, ranap_msg);
+	msgb_enqueue(&map->sccp_fi_ctx.wait_cc_tx_msg_list, ranap_msg);
+}
 
-		kpi_ranap_process_dl(map, message);
-
-		if (!map->is_ps) {
-			/* Circuit-Switched. Set up mapping of RTP ports via MGW */
-
-			switch (message->procedureCode) {
-			case RANAP_ProcedureCode_id_RAB_Assignment:
-				/* mgw_fsm_alloc_and_handle_rab_ass_req() takes ownership of (ranap) message */
-				return handle_cs_rab_ass_req(map, ranap_msg, message);
-			case RANAP_ProcedureCode_id_Iu_Release:
-				/* Any IU Release will terminate the MGW FSM, the message itsself is not passed to the
-				 * FSM code. It is just forwarded normally by map_rua_tx_dt() below. */
-				mgw_fsm_release(map);
-				break;
-			}
-#if ENABLE_PFCP
-		} else {
-			switch (message->procedureCode) {
-			case RANAP_ProcedureCode_id_RAB_Assignment:
-				/* If a UPF is configured, handle the RAB Assignment via ps_rab_ass_fsm, and replace the
-				 * GTP F-TEIDs in the RAB Assignment message before passing it on to RUA. */
-				if (hnb_gw_is_gtp_mapping_enabled()) {
-					LOGP(DMAIN, LOGL_DEBUG,
-					     "RAB Assignment: setting up GTP tunnel mapping via UPF %s\n",
-					     osmo_sockaddr_to_str_c(OTC_SELECT, osmo_pfcp_cp_peer_get_remote_addr(g_hnbgw->pfcp.cp_peer)));
-					return hnbgw_gtpmap_rx_rab_ass_req(map, ranap_msg, message);
-				}
-				/* If no UPF is configured, directly forward the message as-is (no GTP mapping). */
-				LOGP(DMAIN, LOGL_DEBUG, "RAB Assignment: no UPF configured, forwarding as-is\n");
-				break;
-
-			case RANAP_ProcedureCode_id_Iu_Release:
-				/* Any IU Release will terminate the MGW FSM, the message itsself is not passed to the
-				 * FSM code. It is just forwarded normally by map_rua_tx_dt() below. */
-				hnbgw_gtpmap_release(map);
-				break;
-			}
-#endif
-		}
-	}
-
-	/* It was not a RAB Assignment Request that needed to be intercepted. Forward as-is to RUA. */
-	return map_rua_dispatch(map, MAP_RUA_EV_TX_DIRECT_TRANSFER, ranap_msg);
+static struct msgb *wait_cc_tx_msg_list_dequeue(struct hnbgw_context_map *map)
+{
+	struct msgb *ranap_msg = msgb_dequeue(&map->sccp_fi_ctx.wait_cc_tx_msg_list);
+	if (ranap_msg)
+		talloc_steal(OTC_SELECT, ranap_msg);
+	return ranap_msg;
 }
 
 static void map_sccp_init_action(struct osmo_fsm_inst *fi, uint32_t event, void *data)
 {
-	struct hnbgw_context_map *map = fi->priv;
 	struct msgb *ranap_msg = NULL;
 
 	switch (event) {
@@ -280,10 +225,6 @@ static void map_sccp_init_action(struct osmo_fsm_inst *fi, uint32_t event, void
 	case MAP_SCCP_EV_RAN_LINK_LOST:
 	case MAP_SCCP_EV_USER_ABORT:
 	case MAP_SCCP_EV_CN_LINK_LOST:
-		ranap_msg = data;
-		/* No CR has been sent yet, just go to disconnected state. */
-		if (msg_has_l2_data(ranap_msg))
-			LOG_MAP(map, DLSCCP, LOGL_ERROR, "SCCP not connected, cannot dispatch RANAP message\n");
 		map_sccp_fsm_state_chg(MAP_SCCP_ST_DISCONNECTED);
 		return;
 
@@ -320,17 +261,14 @@ static void map_sccp_wait_cc_action(struct osmo_fsm_inst *fi, uint32_t event, vo
 		return;
 
 	case MAP_SCCP_EV_TX_DATA_REQUEST:
-		/* ranap_msg = data; */
-		LOGPFSML(fi, LOGL_ERROR, "Connection not yet confirmed, cannot forward RANAP to CN\n");
+		ranap_msg = data;
+		LOGPFSML(fi, LOGL_INFO, "Caching RANAP msg from RUA while waiting for SCCP CC\n");
+		wait_cc_tx_msg_list_enqueue(map, ranap_msg);
 		return;
 
 	case MAP_SCCP_EV_RAN_LINK_LOST:
 	case MAP_SCCP_EV_USER_ABORT:
 	case MAP_SCCP_EV_CN_LINK_LOST:
-		ranap_msg = data;
-		/* RUA connection was terminated. First wait for the CC before releasing the SCCP conn. */
-		if (msg_has_l2_data(ranap_msg))
-			LOGPFSML(fi, LOGL_ERROR, "Connection not yet confirmed, cannot forward RANAP to CN\n");
 		map->please_disconnect = true;
 		return;
 
@@ -357,9 +295,15 @@ static void map_sccp_wait_cc_action(struct osmo_fsm_inst *fi, uint32_t event, vo
 static void map_sccp_connected_onenter(struct osmo_fsm_inst *fi, uint32_t prev_state)
 {
 	struct hnbgw_context_map *map = fi->priv;
+	struct msgb *ranap_msg;
+
+	/* Now that SCCP conn is confirmed, forward pending msgs received from RUA side: */
+	while ((ranap_msg = wait_cc_tx_msg_list_dequeue(map)))
+		tx_sccp_df1(fi, ranap_msg);
+
 	if (map->please_disconnect) {
-		/* SCCP has already been asked to disconnect, so disconnect now that the CC has been received. Send RLSD
-		 * to SCCP (without RANAP data) */
+		/* SCCP has already been asked to disconnect, so disconnect now that the
+		 * CC has been received. Send RLSD to SCCP (without RANAP data) */
 		tx_sccp_rlsd(fi);
 		map_sccp_fsm_state_chg(MAP_SCCP_ST_DISCONNECTED);
 	}
@@ -409,21 +353,14 @@ static void map_sccp_connected_action(struct osmo_fsm_inst *fi, uint32_t event,
 	case MAP_SCCP_EV_USER_ABORT:
 		/* The user is asking for disconnection, so there is no Iu Release in progress. Disconnect now. */
 	case MAP_SCCP_EV_CN_LINK_LOST:
-		ranap_msg = data;
 		/* The CN peer has sent a RANAP RESET, so the old link that this map ran on is lost */
-
-		/* There won't be any ranap_msg, but if a caller wants to dispatch a msg, forward it before
-		 * disconnecting. */
-		tx_sccp_df1(fi, ranap_msg);
 		tx_sccp_rlsd(fi);
 		map_sccp_fsm_state_chg(MAP_SCCP_ST_DISCONNECTED);
 		return;
 
 	case MAP_SCCP_EV_RX_RELEASED:
 		ranap_msg = data;
-		/* The CN sends an N-Disconnect (SCCP Released) out of the usual sequence. Not what we expected, but
-		 * handle it. */
-		LOGPFSML(fi, LOGL_ERROR, "CN sends SCCP Released sooner than expected\n");
+		/* The CN sends an N-Disconnect (SCCP Released). */
 		handle_rx_sccp(fi, ranap_msg);
 		map_sccp_fsm_state_chg(MAP_SCCP_ST_DISCONNECTED);
 		return;
@@ -552,6 +489,7 @@ void map_sccp_fsm_cleanup(struct osmo_fsm_inst *fi, enum osmo_fsm_term_cause cau
 {
 	struct hnbgw_context_map *map = fi->priv;
 	map->sccp_fi = NULL;
+	msgb_queue_free(&map->sccp_fi_ctx.wait_cc_tx_msg_list);
 }
 
 #define S(x)    (1 << (x))
